/* $Id: omega_flats.c,v 1.2 2012-01-31 13:38:50 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012-01-31 13:38:50 $
 * $Revision: 1.2 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>


#include "omega_flats.h"
#include "omega_fftw.h"
#include "omega_dfs.h"
#include "omega_pfits.h"
#include "omega_utils.h"
#include "omega_stats.h"

/**
 * @defgroup  omega_flats    Flat Fields
 *
 * @par Synopsis
 *   This module contains functions to create a flat field (master flat, night sky flat, fringes flat). It
 *   also has functions to correct a science frame of a flat.
 *
 * @code
 *   #include omega_flat.h
 *   #include omega_fftw.h
 *   #include omega_dfs.h
 *   #include omega_pfits.h
 *   #include omega_utils.h
 *   #include omega_stats.h
 * @endcode
 */
/**@{*/

/**
 * @brief		It creates a master flat field
 * @param dome		Dome flat image
 * @param twil		Twilight flat image
 * @param bpmmap	Bad pixels map
 * @param pars		Parameters list
 *
 * @return	twil_comp	A master flat field or NULL in case of error
 */
cpl_image *omega_make_mflat(cpl_image *dome, cpl_image *twil,
		cpl_mask *bpmmap, cpl_parameterlist *pars)
{

	int dx = 0;
	int dy = 0;

	double gausfilt = 9.0;

	cpl_image 	*dome_comp;
	cpl_image 	*dome_low_image;
	cpl_image 	*twil_comp;
	cpl_parameter *par;

	if((dome == NULL) || (twil == NULL) || (pars == NULL))
		return NULL;


	/* Retrieve input parameters */
	par = cpl_parameterlist_find(pars, "omega.omega_mflat.GaussianFiltSize") ;
	gausfilt = cpl_parameter_get_double(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_mflat.MirrorXpix") ;
	dx = cpl_parameter_get_int(par) ;

	par = cpl_parameterlist_find(pars, "omega.omega_mflat.MirrorYpix") ;
	dy = cpl_parameter_get_int(par) ;

	/* Get low spatial frequency from twilight flat */
	cpl_msg_info (cpl_func,"Working on low spatial frequency component of flat.");
	twil_comp = omega_get_spatial_freq(twil, bpmmap, gausfilt, dx, dy);
	/*	twil_comp = cpl_image_duplicate(twil);*/

	if (twil_comp == NULL) {
		cpl_msg_error (cpl_func, "Error in low spatial frequency calculation");
		return NULL;
	}

	/*Get high spatial frequency from dome flat */
	cpl_msg_info (cpl_func,"Working on high spatial frequency component of flat");
	dome_low_image = omega_get_spatial_freq(dome, bpmmap, gausfilt, dx, dy);
	if (dome_low_image == NULL) {
		cpl_msg_error (cpl_func, "Error in high spatial frequency calculation");
		freeimage(twil_comp);
		return NULL;
	}

	dome_comp = cpl_image_divide_create(dome, dome_low_image);
	if (dome_comp == NULL) {
		cpl_msg_error(cpl_func,"Error in division. <%s>",cpl_error_get_message());
		freeimage(twil_comp);
		freeimage(dome_low_image);
		return NULL;
	}
	freeimage(dome_low_image);

	/*Combine the two contributions*/
	cpl_image_multiply(twil_comp, dome_comp);
	freeimage(dome_comp);

	if(bpmmap != NULL)
		cpl_image_reject_from_mask(twil_comp, bpmmap);


	/* Normalize the image */
	if (cpl_image_normalise(twil_comp, CPL_NORM_MEAN) != CPL_ERROR_NONE){
		cpl_msg_error (cpl_func, "Image is NULL in normalization step. <%s>",cpl_error_get_message());
		freeimage(twil_comp);
		return NULL;
	}

	return twil_comp;

}


/**
* @brief    This function takes a reduced science image and correct from
*			the fringing.
*
* @param    sci      	Reduced Science Image to correct
* @param    fringe     Fringe Image
* @param    bpm	    Bad Pixels Map
* @param	 fpars		Parameters List
*
* @return	0 on success or -1 in case of error.
*/

int omega_fringecor(cpl_image *sci, const cpl_image *fringe, const cpl_image *bpm,
		const cpl_parameterlist *fpars)
{

	int i = 0;
	int N = 0;
	int npars = 0;
	double a = 0.0;
	double FF = 0.0;
	double hthre = 5.0;
	double lthre = 1.5;
	double mean_F = 0.0;
	double mean_R = 0.0;
	double median = 0.0;
	double RF = 0.0;
	double scale = 5.0;
	double stdev = 0.0;
	const char *text = NULL;

	const cpl_parameter 	*par;
	cpl_stats 		*stats;
	cpl_stats 		*stats_F;
	cpl_stats 		*stats_R;
	cpl_image 		*image;
	cpl_image		*ff;
	//    cpl_image 		*fringed;
	cpl_mask 		*map, *bpmap;
	cpl_mask 		*bin1;
	cpl_mask 		*bin2;
	cpl_error_code 	 code;

	if((sci == NULL) || (fringe == NULL) || (bpm == NULL) || (fpars == NULL))
		return -1;

	/*Retrieve fringing parameters*/
	par = cpl_parameterlist_get_first_const(fpars);

	for(i=0; i<npars; i++) {
		text = cpl_parameter_get_alias(par,CPL_PARAMETER_MODE_CLI);
		if(strcmp("sigma-fringes", text) == 0) {
			scale = cpl_parameter_get_double(par) ;
		}
		else if(strcmp("lfringes", text) == 0) {
			lthre = cpl_parameter_get_double(par) ;
		}
		else if(strcmp("hfringes", text) == 0) {
			hthre = cpl_parameter_get_double(par) ;
		}

		par = cpl_parameterlist_get_next_const(fpars);

	}

	/* Get statistics of science image */
	stats = cpl_stats_new_from_image(sci, CPL_STATS_ALL);
	if(stats == NULL)
		return -1;

	median = cpl_stats_get_median(stats);
	stdev = cpl_stats_get_stdev(stats);

	/*it sounds like & is a bitwise AND*/
	bpmap = cpl_mask_threshold_image_create(bpm, 0.5, 1.5);
	map = cpl_mask_threshold_image_create(sci,median-(scale*stdev),median+(scale*stdev));
	cpl_mask_and(bpmap, map);
	cpl_mask_delete(map);
	freestats(stats);


	/* Get statistics of fringe image */
	ff = cpl_image_duplicate(fringe);
	stats = cpl_stats_new_from_image(ff, CPL_STATS_ALL);
	if(stats == NULL)
		return -1;

	median = cpl_stats_get_median(stats);
	stdev = cpl_stats_get_stdev(stats);

	/*Do an OR then an AND for next step*/
	bin1 = cpl_mask_threshold_image_create(ff, median-(hthre*stdev), median-(lthre*stdev));
	bin2 = cpl_mask_threshold_image_create(ff, median+(lthre*stdev), median+(hthre*stdev));

	code = cpl_mask_or(bin1, bin2);
	if(code != CPL_ERROR_NONE) {
		cpl_msg_debug(cpl_func,"Error in bit-wise OR operation");
		freemask(bin1);
		freemask(bin2);
		freemask(bpmap);
		freestats(stats);
		return -1;
	}

	code = cpl_mask_and(bpmap, bin1);
	if(code != CPL_ERROR_NONE) {
		cpl_msg_debug(cpl_func,"Error in bit-wise AND operation");
		freemask(bin1);
		freemask(bin2);
		freestats(stats);
		return -1;
	}

	freemask(bin1);
	freemask(bin2);
	freestats(stats);

	/*
	 * Calculate statistics only on good pixels, on the two images
	 */

	/* Sets bad pixels in the fringe image as defined in mask */
	code = cpl_image_reject_from_mask(ff, bpmap);
	if(code != CPL_ERROR_NONE)
		cpl_msg_warning(cpl_func,"Cannot set BP in fringe image");

	stats_F = cpl_stats_new_from_image(ff, CPL_STATS_ALL);
	mean_F = cpl_stats_get_mean(stats_F);

	/* Sets bad pixels in the image as defined in mask */
	code = cpl_image_reject_from_mask(sci, bpmap);
	if(code != CPL_ERROR_NONE)
		cpl_msg_warning(cpl_func,"Cannot set BP in image");

	stats_R = cpl_stats_new_from_image(sci, CPL_STATS_ALL);
	mean_R = cpl_stats_get_mean(stats_R);

	N = cpl_stats_get_npix(stats_F);

	image = cpl_image_multiply_create(sci, ff);
	stats = cpl_stats_new_from_image(image, CPL_STATS_ALL);
	RF = cpl_stats_get_flux(stats);

	freeimage(image);
	freestats(stats);

	image = cpl_image_multiply_create(ff, ff);
	stats = cpl_stats_new_from_image(image, CPL_STATS_ALL);
	FF = cpl_stats_get_flux(stats);

	a = (RF - N*mean_R*mean_F) / (FF - N*mean_F*mean_F);
	/*	cpl_msg_info("","The measured scale factor is %f", a);*/

	cpl_image_multiply_scalar(ff, a);

	cpl_image_subtract(sci, ff);

	/*Clean up*/

	freeimage(ff);
	freeimage(image);
	freestats(stats);
	freestats(stats_F);
	freestats(stats_R);

	return 0;
}

/**
* @brief		Divide image by flat field
*
* @param	image	Image to divide flat from
* @param	flat	Flat field image
* @param	nksy	(optional) night sky flat
*
* @return	0 on success or -1 in case of error
*/
int omega_flatcor(cpl_image *image, const cpl_image *flat, const cpl_image *nsky)
{
	cpl_image *mflat;

	if((image == NULL) || (flat == NULL))
		return -1;

	/* If there is a night sky flat, use it. */
	if(nsky != NULL){
		mflat = cpl_image_multiply_create(flat, nsky);
		cpl_image_divide(image, mflat);
		freeimage(mflat);
	}
	else {
		cpl_image_divide(image, flat);
	}

	return 0;

}

/**
*
* @brief		Create a night sky flat from a series of science frames
*
* @param	ilist	Image list with reduced science frames
* @param	bpm		Optional bad pixels image
* @param	n		Number of images in list
*
* @return	nsky	Night sky image or NULL in case of error.
*
*/
cpl_image *omega_nsky_create(cpl_imagelist *ilist,cpl_image *bpm, int n)
{

	int i;
	cpl_image *nsky;

	if(ilist == NULL)
		return NULL;

	for(i = 0; i < n; i++){
		cpl_image *img = cpl_imagelist_get(ilist, i);
		if(bpm != NULL){
			cpl_mask *mask = cpl_mask_threshold_image_create(img, 0.5, 1.5);
			cpl_image_reject_from_mask(img, mask);
			freemask(mask);
		}
		double median = cpl_image_get_median(img);
		cpl_image_divide_scalar(img, median);
	}

	/* Create median image */
	nsky = cpl_imagelist_collapse_median_create(ilist);

	/* Normalize it */
	cpl_image_normalise(nsky,CPL_NORM_MEAN);

	if(bpm != NULL)
		cpl_image_multiply(nsky, bpm);

	return nsky;
}

/**
*
* @brief		Create a fringe image from a list of science frames
*
* @param	ilist	Image list with science images
* @param	bpm		Optional bad pixels image
* @param	n		Number of images
*
* @return	fringes		Image of fringes
*
*/
cpl_image *omega_fringes_create(cpl_imagelist *ilist, cpl_image *bpm, int n)
{
	cpl_image *fringes;

	if(ilist == NULL)
		return NULL;


	fringes = omega_nsky_create(ilist, bpm, n);
	cpl_image_subtract_scalar(fringes, 1.0);

	return fringes;

}

/**@}*/
