/* $Id: omega_science.c,v 1.39 2012-08-22 09:51:58 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012-08-22 09:51:58 $
 * $Revision: 1.39 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include "omega_recipe.h"

#include "omega_coadd.h"
#include "omega_catalog.h"
#include "omega_background.h"
#include "omega_cosmic.h"
#include "omega_photometry.h"
#include "omega_science.h"
#include "omega_satellites.h"
#include "omega_wcscor.h"
/**
 * @defgroup omega_science  Science
 *
 * @par Synopsis:
 * This recipe is used to reduced a series of science frames . It will identify the frame among the different modes: \n
 * STARE, JITTER, DITHER or OFFSET. The common steps applied to any science frame are: trimming, correcting for \n
 * the overscan (if requested), correcting for the bias and the flat. The individual science frames are also \n
 * matched against the USNOA2 catalogues to derive a WCS correction. The zeropoints table is taken into \n
 * account to refine the zeropoint of the science image. Each science image is saved as a product. The next step\n
 * is to create a stack of the individual frames. The combination mode can be modified in the input parameter\n
 * <i>comb-mode</i>. The combination step will also produce a combination map which is saved as a product. The final\n
 * stack image is matched once more with the USNOA2 catalogues to obtain a more refined WCS correction. A\n
 * catalogue is created from the stack image and also saved as a product. The recipe expects the following \n
 * tags in the Set of Frames:
 *
 * @par Input frames:\n
 * STARE, JITTER, DITHER or OFFSET\n
 * MASTER_BIAS\n
 * MASTER_FLAT\n
 * BAD_PIXELS_MAP\n
 * ZEROPOINTS\n
 * MASTER_USNOA2_CATALOGUE\n
 * MASTER_NIGHTSKY_FLAT (optional)\n
 * MASTER_FRINGES_FLAT (optional)\n
 * ILLUMINATION (optional)
 *
 * @par Output frames:\n
 * SIMPLE_SCIENCE\n
 * STACK_SCIENCE\n
 * CONF_MAP_SCIENCE\n
 * SCIENCE_CATALOGUE
 *
 *
 * @code
 *      #include omega_recipe.h
 *      #include omega_background.h
 *      #include omega_coadd.h
 *      #include omega_catalog.h
 *      #include omega_cosmic.h
 *      #include omega_photometry.h
 *      #include omega_satellites.h
 *      #include omega_science.h
 *      #include omega_wcscor.h
 *
 * @endcode
 */
/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_science_create(cpl_plugin *) ;
static int omega_science_exec(cpl_plugin *) ;
static int omega_science_destroy(cpl_plugin *) ;
static int omega_science(cpl_frameset *,cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/

static int omega_science_retrieve_input_param(const cpl_parameterlist  *parlist);
static int omega_science_load_calib();
static void omega_science_init(void);
static void omega_science_tidy(int level);
static cpl_error_code omega_correct_crpix_union(cpl_propertylist *xlist,
		cpl_bivector *offsets);


/**
 *
 * @brief
 *   Some of the input parameters for the science recipe
 */
static struct {
    int extnum;
    int oc;
    int sub_back;
    int mode;

}omega_science_config;

/**
 *
 * @brief
 *   The QC parameters calculated for the science recipe
 */
static struct {
    int nmatches;
    int nstars;
    double mratio;
    float ZP;
    float ZPerr;
//    float Extinction;
//    float Exterr;
    double Seeing;

}omega_sci_qc;

/*
 * Input and Ouput
 */
static struct {

    /* Calib frames */
    const cpl_frame *bpmfr;
    const cpl_frame *fringesfr;
    const cpl_frame *illumfr;
    const cpl_frame *mbiasfr;
    const cpl_frame *mflatfr;
    const cpl_frame *nskyfr;
    const cpl_frame *usnoa2;
    const cpl_frame *photomfr;

    cpl_size         *labels;
    cpl_frameset     *scilist;
    cpl_propertylist *ph;
    cpl_propertylist *eh;
    omega_fits        *scifits;
    cpl_image        *mbias;
    cpl_image        *mflat;
    cpl_image        *bpm;
    cpl_image        *illum;
    cpl_image        *fringes;
    cpl_image        *nsky;
    cpl_table        *photom;

    /* Products */
    char *proname;
    char *confname;
    char *catname;
    omega_fits        *sfits;
    cpl_image        **combined;
    cpl_table        *catalogue;
    cpl_propertylist *qclist;
    cpl_propertylist *plist;
}ps;


/* Static variables */
#define RECIPE "omega_science"


/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
	cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
	cpl_plugin  *   plugin = &recipe->interface ;

	cpl_plugin_init(plugin,
			CPL_PLUGIN_API,
			OMEGA_BINARY_VERSION,
			CPL_PLUGIN_TYPE_RECIPE,
			"omega_science",
			"OMEGA - This recipe is used to reduce science observations.",
			"The recipe assumes that the data is for a single CCD. \n"
			"The following operations are performed: \n\n"
			" 	 1. Trim and overscan correct; \n"
			"    2. Subtract the bias and divide by the flat field; \n"
			"    3. Correct for fringing, illumination and the background, if requested; \n"
			" 	 4. Extract the zero point from the data and calculate the extinction \n"
			"	    value and save both in the header and in one table. \n\n"
			" ------------------------------------------------------------------------ \n\n"
			" Mandatory inputs : \n\n"
			"	: a raw standard star frame \n"
			"	: a master bias frame (Calfile 541) \n"
			"	: a master flat field (Calfile 546) \n\n"
			"   : a master USNOA2 table \n"
			"   : a standard stars catalog table \n\n"
			" Optional inputs : \n\n"
			"	: fringe map (Calfile 545) \n"
			"	: bad pixels map (Calfile 522+535) \n"
			"	: illumination correction frame (Calfile 548) \n\n"
			" The recipe will save the extinction and zero point values in one table.",
			"Sandra Castro / Armin Gabasch",
			"scastro@eso.org",
			omega_get_license(),
			omega_science_create,
			omega_science_exec,
			omega_science_destroy) ;

	cpl_pluginlist_append(list, plugin) ;

	return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
static int omega_science_create(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	cpl_parameter * p ;
	char *path = NULL;

	/* Do not create the recipe if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Create the parameters list in the cpl_recipe object */
	recipe->parameters = cpl_parameterlist_new() ;
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Parameter list allocation failed");
		cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
	}

	/* Fill the parameters list */

	/* --------------- General parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_science.ExtensionNumber",
			CPL_TYPE_INT,
			"FITS extension number to load (1 to 32). (-1 == all)",
			"omega_science",
			-1) ;

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_range("omega.omega_science.OverscanMethod",
			CPL_TYPE_INT,
			"Overscan Correction Method",
			"overscan",
			0, 0, 6);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	/* --------------- Saturated pixels parameters ---------------------------*/

	p = cpl_parameter_new_value("omega.omega_science.LowThreSatuPixel",
			CPL_TYPE_DOUBLE,
			"Low threshold for calculating the saturated pixels map.",
			"omega_science",
			-500.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "lt-satu") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_science.HighThreSatuPixel",
			CPL_TYPE_DOUBLE,
			"High threshold for calculating the saturated pixels map.",
			"omega_science",
			60000.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "ht-satu") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	/* --------------- Satellite detection parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_science.DetectionThreSatellite",
			CPL_TYPE_DOUBLE,
			"Minimum SNR for pixels to contribute to Hough map.",
			"omega.Hough",
			4.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "det-sate") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_science.HoughThreshold",
			CPL_TYPE_DOUBLE,
			"Threshold for satellite tracks in Hough image.",
			"omega.Hough",
			400.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "hough-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* --------------- Fringing  parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_science.SigmaFringeScaling",
			CPL_TYPE_DOUBLE,
			"Sigma threshold in image data for scaling estimate (fringes only)",
			"omega_science",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sigma-fringes") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_science.LowThreFringe",
			CPL_TYPE_DOUBLE,
			"Lower bound of fringes to include in scaling (fringes only)",
			"omega_science",
			1.5) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "lfringes") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_science.HighThreFringe",
			CPL_TYPE_DOUBLE,
			"Higher bound of fringes to include in scaling (fringes only)",
			"omega_science",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "hfringes") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* ----------------- Background  parameters ---------------------------*/
	p = cpl_parameter_new_value("omega.omega_science.SubtractBackground",
			CPL_TYPE_BOOL,
			"Boolean value to subtract background or not. 1(TRUE), 0(FALSE)",
			"omega_science",
			1) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sub-back") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_science.BackThreshold",
			CPL_TYPE_DOUBLE,
			"Detection threshold for background in image (Sextractor DETECT_THRESH)",
			"omega_science",
			1000.) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "backthre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* ------------------ PATH to external executable programs --------------*/
	path = cpl_sprintf("%s", OMEGA_BIN_PATH);
	p = cpl_parameter_new_value("omega.omega_science.BinPath",
			CPL_TYPE_STRING,
			"Path to any external executable program.",
			"omega.BinPath",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "bin-path");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	/* ------------------ Sextractor parameters ------------------------- */
	path = cpl_sprintf("%s/omega.sex", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_science.SexConfig",
			CPL_TYPE_STRING,
			"Path to Sextractor config file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-config");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);


	path = cpl_sprintf("%s/omega.conv", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_science.SexConv",
			CPL_TYPE_STRING,
			"Path to Sextractor convolution mask file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-conv");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);


	path = cpl_sprintf("%s/omega.param", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_science.SexParam",
			CPL_TYPE_STRING,
			"Path to Sextractor parameters file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-param");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	path = cpl_sprintf("%s/omega.nnw", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_science.SexNnw",
			CPL_TYPE_STRING,
			"Path to Sextractor neural network config file.",
			"omega.sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-nnw");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	p = cpl_parameter_new_value("omega.omega_science.SexDetThreshold",
			CPL_TYPE_DOUBLE,
			"Detection threshold for Sextractor sources.",
			"omega.sextractor",
			1.5);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-dthre");
	cpl_parameterlist_append(recipe->parameters, p);


	/* ------- Cosmic rays detection parameters for Sextractor ------------*/
	path = cpl_sprintf("%s/omega.cosmic.sex", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_science.SexCosmic",
			CPL_TYPE_STRING,
			"Path to Sextractor cosmic detection mode config file.",
			"omega.Sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-cosmic");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	path = cpl_sprintf("%s/omega.cosmic.param", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_science.SexCosmicParam",
			CPL_TYPE_STRING,
			"Path to Sextractor cosmic parameters file.",
			"omega.Sextractor",
			path);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"sex-cosmic-param");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	path = cpl_sprintf("%s/cosmic.ret", OMEGA_CONFIG_PATH);
	p = cpl_parameter_new_value("omega.omega_science.SexCosmicFilt",
			CPL_TYPE_STRING,
			"Path to Sextractor filter mask for cosmic detection mode.",
			"omega.Sextractor",
			path);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-cosfilt");
	cpl_parameterlist_append(recipe->parameters, p);
	cpl_free(path);

	p = cpl_parameter_new_value("omega.omega_science.SexCosmicDet",
			CPL_TYPE_DOUBLE,
			"Detection threshold for detecting cosmic rays in Sextractor",
			"omega.Sextractor",
			5.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-cosmic-det") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* ------------------ Astrometric parameters ------------------------- */
	p = cpl_parameter_new_value("omega.omega_science.AstromDetThre",
			CPL_TYPE_DOUBLE,
			"Sextractor threshold to detect stars for astrometric correction.",
			"omega.Astrom",
			10.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "astrom-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_enum("omega.omega_science.PlateSolution",
			CPL_TYPE_INT,
			"Number of constant plate fits to do for the astrometry solution.",
			"omega.Astrom",
			4,2,4,6);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "platesol");
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_science.NumIterations",
			CPL_TYPE_INT,
			"Number of iterations for the plate solution fitting.",
			"omega.Astrom",
			3);

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "niter");
	cpl_parameterlist_append(recipe->parameters, p);


	p = cpl_parameter_new_range("omega.omega_science.SextractorFlagLimit",
			CPL_TYPE_INT,
			"Sextractor FLAG upper limit to preselect good sources for the "
			"astrometric correction computation.", "omega.Astrom",
			256, 0, 256);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"SextractorFlagLimit") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_range("omega.omega_science.SextractorClassStar",
			CPL_TYPE_DOUBLE,
			"Sextractor CLASS_STAR lower limit to preselect starlike objects "
			"for the astrometric correction computation.", "omega.Astrom",
			0.0, 0.0, 1.0);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"SextractorClassStar") ;
	cpl_parameterlist_append(recipe->parameters, p) ;



	/* ------------------ Photometric parameters ------------------------- */
	p = cpl_parameter_new_value("omega.omega_science.PhotomDetThre",
			CPL_TYPE_DOUBLE,
			"Sextractor threshold to detect stars for photometric correction.",
			"omega.Photom",
			10.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "photom-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_science.MatchPhotomRadius",
			CPL_TYPE_DOUBLE,
			"Radius within which to match photometric stds and stars",
			"omega.Photom",
			10.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "radius") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	p = cpl_parameter_new_value("omega.omega_science.PhotomAperture",
			CPL_TYPE_DOUBLE,
			"Aperture in pixels for photometric measurements in Sextractor",
			"omega.Photom",
			30.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "aperture") ;
	cpl_parameterlist_append(recipe->parameters, p) ;

	/* ------------------ Pattern Matching parameters ------------------------- */
	p = cpl_parameter_new_value("omega.omega_science.Nstars",
			CPL_TYPE_INT,
			"Number of stars in image to use in "
			"pattern matching",
			"omega.Pattern",
			25);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "nstars");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);


	p = cpl_parameter_new_value("omega.omega_science.Npattern",
			CPL_TYPE_INT,
			"Number of catalogue sources to use in "
			"pattern matching",
			"omega.Pattern",
			15);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "npattern");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);


	p = cpl_parameter_new_value("omega.omega_science.Tolerance",
			CPL_TYPE_DOUBLE,
			"Max relative difference of angles and scales from their "
			"median value for match acceptance (tolerance)",
			"omega.Pattern",
			0.1);

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "tol");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_science.SearchPatternRadius",
			CPL_TYPE_DOUBLE,
			"Search radius for full PPM (pixels)",
			"omega.Pattern",
			10.0);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "pradius");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_science.MaxPatternRadius",
			CPL_TYPE_DOUBLE,
			"Maximum search radius for full PPM (pixels)",
			"omega.Pattern",
			40.0);
	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "maxpradius");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	/* Stacking parameters */
	p = cpl_parameter_new_value("omega.omega_science.xcorr",
			CPL_TYPE_STRING,
			"String of four values meaning: x,y search area and "
			"x,y measurement area, needed for cross-correlation "
			"when stacking images",
			"omega.Stack",
			"20,20,20,20");

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "xcorr");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_science.CombMode",
			CPL_TYPE_STRING,
			"Combination mode to use when stacking images. "
			"union,intersect or first",
			"omega.Stack",
			"union");

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "comb-mode");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	p = cpl_parameter_new_value("omega.omega_science.StackRej",
			CPL_TYPE_STRING,
			"Number of low and high values to reject when stacking",
			"omega.Stack",
			"0,0");

	cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "stack-rej");
	cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
	cpl_parameterlist_append(recipe->parameters, p);

	/* Catalogue creation */
	p = cpl_parameter_new_value("omega.omega_science.CatThre",
			CPL_TYPE_DOUBLE,
			"Sextractor threshold to detect sources in stack image",
			"omega.Sextractor",
			3.0) ;

	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "cat-thre") ;
	cpl_parameterlist_append(recipe->parameters, p) ;



	p = cpl_parameter_new_value("omega.omega_science.mag_zpt_no_std_table",
			CPL_TYPE_DOUBLE,
			"Zeropoint used if the standard star zeropoint table is empty ",
			"omega.Sextractor",
			0.0) ;
	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "mag_zpt_no_std_table") ;
	cpl_parameterlist_append(recipe->parameters, p) ;



	p = cpl_parameter_new_value("omega.omega_science.mag_zpt_err_no_std_table",
			CPL_TYPE_DOUBLE,
			"Zeropoint error used if the standard star zeropoint table is empty ",
			"omega.Sextractor",
			1.0) ;
	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "mag_zpt_err_no_std_table") ;
	cpl_parameterlist_append(recipe->parameters, p) ;



	p = cpl_parameter_new_value("omega.omega_science.ext_no_std_table",
			CPL_TYPE_DOUBLE,
			"extinction used if the standard star zeropoint table is empty",
			"omega.Sextractor",
			1.0) ;
	cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "ext_no_std_table") ;
	cpl_parameterlist_append(recipe->parameters, p) ;


	/* Return */
	return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_science_exec(cpl_plugin * plugin)
{
	cpl_recipe * recipe;
	int recipe_status;

	/* Return immediately if an error code is already set */
	if (cpl_error_get_code() != CPL_ERROR_NONE) {
		cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
				cpl_func, __LINE__, cpl_error_get_where());
		return (int)cpl_error_get_code();
	}

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	/* Verify parameter and frame lists */
	if (recipe->parameters == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}
	if (recipe->frames == NULL) {
		cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Invoke the recipe */
	recipe_status = omega_science(recipe->frames, recipe->parameters);

	/* Ensure DFS-compliance of the products */
	if (cpl_dfs_update_product_header(recipe->frames)) {
		if (!recipe_status) recipe_status = (int)cpl_error_get_code();
	}

	return recipe_status;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_science_destroy(cpl_plugin * plugin)
{
	cpl_recipe  *   recipe;

	if (plugin == NULL) {
		cpl_msg_error(cpl_func, "Null plugin");
		cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
	}

	/* Verify plugin type */
	if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
		cpl_msg_error(cpl_func, "Plugin is not a recipe");
		cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
	}

	/* Get the recipe */
	recipe = (cpl_recipe *)plugin;

	cpl_parameterlist_delete(recipe->parameters) ;

	return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    This recipe searches for SCIENCE frames in frameset
  and does a quick reduction on this image. It then derives the zeropoint for
  it and the extinction table.

  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_science(cpl_frameset *set, cpl_parameterlist *pars)
{
	int j,jst,jfn,i,n_ext;
	cpl_size nlab;
	int nraw = 0;
	int refine = 0;
	static int isfirst = 0;
	cpl_size *plabels = NULL;
	int has_fringes = 0;
	double pixscalex = 0.0;
	double pixscaley = 0.0;
	double Seeing = 0.0;
	double ellipticity = 0.0;
	double zeropoint_final=0.0;
	const char *smoothed = "omega_stacksmooth.fits";
	const char *smoothed_weight = "omega_stacksmooth_weight.fits";
	const char *tag = NULL;
	const char *tag_mask = NULL;
	const char * maskname= NULL;

	cpl_image *smooth;
	cpl_image *image_simple_mask=NULL;
	cpl_frame *sciframe;
	cpl_frame *firstframe=NULL;
	cpl_frame *product_frame=NULL ;
	cpl_frame *product_frame_mask=NULL ;
	cpl_frame *conf_frame;
	cpl_frame *catframe;
	cpl_frameset *proset=NULL;
	cpl_frameset *proset_mask=NULL;
	cpl_bivector *offsets=NULL;
	cpl_propertylist *plist, *xlist, *alist;
	cpl_propertylist *plist_dummy=NULL;
	cpl_propertylist *flist;
	cpl_propertylist *qclist=NULL;
	cpl_parameter *par  = NULL;
	const char    *sval = NULL;


	/*Start the recipe */

	if (pars == NULL) {
		cpl_msg_error (cpl_func, "Parameters list not found");
		return -1;
	}

	if (cpl_frameset_is_empty(set) == 1) {
		cpl_msg_error (cpl_func, "Frameset not found");
		return -1;
	}

	/* Retrieve input parameters and initiates specific parameters lists*/
	if(omega_science_retrieve_input_param(pars) != 0){
		cpl_msg_error(cpl_func,"Cannot retrieve input parameters. %s", cpl_error_get_message());
		return -1;
	}

	/* Identify the RAW and CALIB frames in the input frameset */
	if (oc_dfs_set_groups(set)) {
		cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames") ;
		return -1 ;
	}

	/*Initialized things*/
	omega_science_init();
	omega_sci_qc.Seeing = 0.0;
	omega_sci_qc.mratio = 0.0;
	omega_sci_qc.nmatches = 0;
	omega_sci_qc.nstars = 0;

	/* Verify the frameset contents. */
	if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
			&nlab)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot labelise the input frameset");
		omega_science_tidy(0);
		return -1;
	}

	omega_science_config.mode = 0;

	if ((ps.scilist = omega_frameset_subgroup(set,ps.labels,nlab,
			STARE_RAW)) != NULL) {
		omega_science_config.mode = 1;
		refine = 0;
		tag = STARE_RAW;
	}
	else if ((ps.scilist = omega_frameset_subgroup(set,ps.labels,nlab,
			JITTER_RAW)) != NULL) {
		omega_science_config.mode = 2;
		refine = 0;
		tag = JITTER_RAW;
	}
	else if ((ps.scilist = omega_frameset_subgroup(set,ps.labels,nlab,
			DITHER_RAW)) != NULL) {
		omega_science_config.mode = 3;
		refine = 0;
		tag = DITHER_RAW;
	}
	else if ((ps.scilist = omega_frameset_subgroup(set,ps.labels,nlab,
			OFFSET_RAW)) != NULL) {
		omega_science_config.mode = 4;
		refine = 0;
		tag = OFFSET_RAW;
	}
	else {
		cpl_msg_error(cpl_func,"Cannot find science frames in input frameset");
		omega_science_tidy(0);
		return -1;
	}

	nraw = cpl_frameset_get_size(ps.scilist);
	cpl_msg_info(cpl_func,"There are %d %s frames in frame set",nraw,tag);

	/* Check for mandatory calibration frames */
	/* Master Bias */
	ps.mbiasfr = cpl_frameset_find_const(set, OMEGA_CALIB_BIAS);
	if (ps.mbiasfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set", OMEGA_CALIB_BIAS);
		omega_science_tidy(0);
	}
	cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(ps.mbiasfr));

	/* Master Flat */
	ps.mflatfr = cpl_frameset_find_const(set, OMEGA_CALIB_FLAT);
	if(ps.mflatfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set", OMEGA_CALIB_FLAT);
		omega_science_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s", OMEGA_CALIB_FLAT, cpl_frame_get_filename(ps.mflatfr));

	/* Photometric table */
	ps.photomfr = cpl_frameset_find (set,OMEGA_CALIB_ZP);
	if(ps.photomfr == NULL) {
		cpl_msg_warning(cpl_func,"Cannot find %s in frame set", OMEGA_CALIB_ZP);
		//omega_science_tidy(0);
		//return -1;
	}
	else{
		cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_ZP,cpl_frame_get_filename(ps.photomfr));
	}
	/* USNOA2 master table */
	ps.usnoa2 = cpl_frameset_find (set, OMEGA_USNOA2);
	if (ps.usnoa2 == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s frame in frame set",OMEGA_USNOA2);
		omega_science_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s", OMEGA_USNOA2, cpl_frame_get_filename(ps.usnoa2));

	/* Bad Pixels Map */
	ps.bpmfr = cpl_frameset_find (set,OMEGA_CALIB_BPM);
	if(ps.bpmfr == NULL) {
		cpl_msg_error(cpl_func,"Cannot find %s in frame set", OMEGA_CALIB_BPM);
		omega_science_tidy(0);
		return -1;
	}
	cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_BPM,cpl_frame_get_filename(ps.bpmfr));

	/* Optional Calibration frames */
	/* Illumination frame */
	ps.illumfr = cpl_frameset_find(set, OMEGA_CALIB_ILLUM);
	if(ps.illumfr != NULL)
		cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_ILLUM,cpl_frame_get_filename(ps.illumfr));

	/*Night sky flat */
/*	ps.nskyfr = cpl_frameset_find(set, OMEGA_CALIB_NSKY);
	if(ps.nskyfr != NULL)
		cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_NSKY,cpl_frame_get_filename(ps.nskyfr));
*/
	/* Fringe flat */
/*	ps.fringesfr = cpl_frameset_find(set,OMEGA_CALIB_FRINGE);
	if(ps.fringesfr != NULL)
		cpl_msg_info(cpl_func,"Using %s %s",OMEGA_CALIB_FRINGE,cpl_frame_get_filename(ps.fringesfr));*/


	sciframe = cpl_frameset_find(set,tag);
	/* Check which image extension(s) to load */
	omega_extensions(sciframe,omega_science_config.extnum,&jst,&jfn);
    if(omega_science_config.extnum == 0){
    	cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_science_config.extnum);
    	omega_science_tidy(0);
    	return -1;
    }


	/* Loop over all science frames */
	for(i=0; i < nraw; i++){
		cpl_msg_info(cpl_func,"Working on %s",cpl_frame_get_filename(sciframe));

		/* Loop through extensions to reduce simple images */
		/* CHANGE LATER ************************************ */

		for (j = jst; j <= jfn; j++) {
			isfirst = (j == jst);

	     	cpl_msg_indent_more();
			cpl_msg_info(cpl_func,".....Working on extension %d.....",j);
	     	cpl_msg_indent_less();

			/* Load calibration frames */
			if(omega_science_load_calib(j) != 0){
				cpl_msg_error(cpl_func,"Cannot load calibration frame(s)");
				omega_science_tidy(0);
				return -1;
			}

			/* Load omega_fits structure for this extension */
			ps.scifits = omega_fits_load(sciframe, CPL_TYPE_FLOAT, j);

			/* Call processing routine. Mark file as dummy if unsuccessful */
			ps.sfits = omega_jitter_process(ps.scifits,ps.mbias,ps.mflat,ps.bpm,ps.photom,
									ps.usnoa2,ps.illum,ps.fringes,ps.nsky, &zeropoint_final ,pars, j);

			if(ps.sfits == NULL){
				cpl_msg_error(cpl_func,"Unable to reduce image");
				omega_science_tidy(0);
				return -1;
			}

			/* Free some space */
			omega_science_tidy(1);

			/* Save this extension with updated extension header!!! */
			/* NOTE: Should these products have the same PROCATG? */
			if(isfirst){
				tag = NULL;
				if(omega_science_config.mode == 1) {
					ps.proname = cpl_sprintf("%s_%s_%d.fits", INSTRUME,SIMPLE_STARE,i);
					product_frame = omega_product_frame(ps.proname, SIMPLE_STARE, CPL_FRAME_TYPE_IMAGE);
					tag = SIMPLE_STARE;
				}
				if(omega_science_config.mode == 2){
					ps.proname = cpl_sprintf("%s_%s_%d.fits", INSTRUME,SIMPLE_JITTER,i);
					product_frame = omega_product_frame(ps.proname, SIMPLE_JITTER, CPL_FRAME_TYPE_IMAGE);
					tag = SIMPLE_JITTER;
				}
				if(omega_science_config.mode == 3){
					ps.proname = cpl_sprintf("%s_%s_%d.fits", INSTRUME,SIMPLE_DITHER,i);
					product_frame = omega_product_frame(ps.proname, SIMPLE_DITHER, CPL_FRAME_TYPE_IMAGE);
					tag = SIMPLE_DITHER;
				}
				if(omega_science_config.mode == 4){
					ps.proname = cpl_sprintf("%s_%s_%d.fits", INSTRUME,SIMPLE_OFFSET,i);
					product_frame = omega_product_frame(ps.proname, SIMPLE_OFFSET, CPL_FRAME_TYPE_IMAGE);
					tag = SIMPLE_OFFSET;
				}
				maskname=cpl_sprintf("%s_%s_%d.fits", INSTRUME,SIMPLE_MASK,i);
				product_frame_mask = omega_product_frame(maskname, SIMPLE_MASK, CPL_FRAME_TYPE_IMAGE);
				tag_mask=SIMPLE_MASK;
			}

			cpl_msg_info(cpl_func,"Saving SIMPLE image");
			if(omega_save_fits(ps.sfits,set,pars,NULL,CPL_BPP_IEEE_FLOAT,ps.proname,RECIPE,
					product_frame,sciframe,isfirst) == -1){
				cpl_msg_error(cpl_func,"Cannot save SIMPLE image. %s",cpl_error_get_message());
				cpl_free(maskname);
				omega_science_tidy(0);
				return -1;
			}

			/*Extracting the mask file from the image*/
			image_simple_mask=cpl_image_new_from_mask(cpl_image_get_bpm(ps.sfits->image));
			cpl_image_delete(ps.sfits->image); ps.sfits->image=NULL;
			ps.sfits->image=image_simple_mask;

			cpl_msg_info(cpl_func,"Saving SIMPLE MASK file");
			if(omega_save_fits(ps.sfits,set,pars,NULL,CPL_BPP_8_UNSIGNED,maskname,RECIPE,
					product_frame_mask,sciframe,isfirst) == -1){
				cpl_msg_error(cpl_func,"Cannot save SIMPLE image. %s",cpl_error_get_message());
				cpl_free(maskname);
				omega_science_tidy(0);
				return -1;
			}


			freefits(ps.sfits);
		}

		sciframe = cpl_frameset_find(set, NULL);

		freespace(ps.proname);
		cpl_free(maskname);
		product_frame = NULL;
		product_frame_mask= NULL;
	}

	/* Labelise simple products */
	if ((plabels = cpl_frameset_labelise(set,omega_compare_tags,
			&nlab)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot labelise the output frameset");
		omega_science_tidy(0);
		return -1;
	}

	/* Extract simple products into a new frameset */
	if ((proset = omega_frameset_subgroup(set,plabels,nlab,
			tag)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot create simple frameset");
		freespace(plabels);
		omega_science_tidy(0);
		return -1;
	}

	/* Extract simple mask products into a new frameset */
	if ((proset_mask = omega_frameset_subgroup(set,plabels,nlab,
			tag_mask)) == NULL) {
		cpl_msg_error(cpl_func,"Cannot create simple mask frameset");
		freespace(plabels);
		omega_science_tidy(0);
		return -1;
	}


	freespace(plabels);

	/* Get telescope offsets (in pixels) from headers */


/*
	if((offsets = omega_tel_offsets(proset, omega_science_config.mode)) == NULL){
		cpl_msg_error(cpl_func,"Unable to get telescope offsets for mode=%d",omega_science_config.mode);
		freeframeset(proset);
		omega_science_tidy(0);
		return -1;
	}
*/

/*
	offsets=omega_get_offsets_from_wcs(proset, cpl_frameset_get_first(proset),
			omega_science_config.extnum);
*/

//    cpl_bivector_dump(offsets,stdout);

	/* Loop over extensions again to create stack */
	/* Check if only one extension was requested */
	firstframe = cpl_frameset_get_first(proset);
	n_ext = cpl_fits_count_extensions(cpl_frame_get_filename(firstframe));

	/* Get filter information */
	plist = cpl_propertylist_load_regexp(cpl_frame_get_filename(firstframe),0,"ESO INS FILT",0);
	flist = omega_pfits_get_filter_info(plist);
	has_fringes = cpl_propertylist_get_int(flist, "HAS_FRINGES");
	freeplist(plist);
	freeplist(flist);


	cpl_msg_info(cpl_func,"Creating stack image and catalogue");
	for(j = 1; j <= n_ext; j++){
		isfirst = (j == 1);

		offsets=omega_get_offsets_from_wcs(proset, firstframe, j);


		cpl_msg_indent_more();
		cpl_msg_info(cpl_func,"....... Image in position %d .......",j);
		cpl_msg_indent_less();

		plist = cpl_propertylist_load(cpl_frame_get_filename(firstframe),j);
		omega_get_pixelscale(plist, &pixscalex, &pixscaley);
		freeplist(plist);

		/*This is a workaround as the current cpl version 5.1 has a bug:
		 * Even if there is only one image in the sof (i.e. the relative shift
		 * is zero), cpl_geom_img_offset_combine() interpolates the bad pixels
		 * dilating one bad pixel to 16 bad pixels */
		if(cpl_frameset_get_size(proset)==1){
			ps.combined = cpl_malloc(2*sizeof(cpl_image *));
			ps.combined[0]=cpl_image_load(cpl_frame_get_filename(cpl_frameset_get_first(proset)),CPL_TYPE_FLOAT,0,j);
			ps.combined[1]=cpl_image_load(cpl_frame_get_filename(cpl_frameset_get_first(proset_mask)),CPL_TYPE_INT,0,j);
			cpl_image_threshold(ps.combined[1],0.5,0.6,1,0);
		}
		else{
			 /* Create stack and confidence map */
			ps.combined = omega_jitter_stack(proset, proset_mask, offsets, refine, pars, j);
		}

		if(ps.combined == NULL){
			cpl_msg_error(cpl_func,"Unable to create stack image. %s", cpl_error_get_message());
			freebivector(offsets);
			freeframeset(proset);
			freeframeset(proset_mask);
			omega_science_tidy(0);
			return -1;
		}

		/* Create night sky flat and correct stack frame */
		if(cpl_frameset_get_size(proset) >= 3){
			cpl_imagelist *ilist = cpl_imagelist_load_frameset(proset, CPL_TYPE_FLOAT,1,j);
			if((ps.nsky = omega_nsky_create(ilist, NULL,
							cpl_frameset_get_size(proset))) != NULL){
				cpl_image_divide(ps.combined[0],ps.nsky);
			}

			/* Correct for fringes if filter requires it */
			if(has_fringes == 1){
				cpl_msg_info(cpl_func,"Applying fringing correction");
				ps.bpm = cpl_image_load(cpl_frame_get_filename(ps.bpmfr),CPL_TYPE_FLOAT,0,j);
				ps.fringes = cpl_image_subtract_scalar_create(ps.nsky, 1.0);
				if(ps.fringes != NULL)
					omega_fringecor(ps.combined[0],ps.fringes,ps.bpm,pars);

				freeimage(ps.fringes);
				freeimage(ps.bpm);
			}
			freeilist(ilist);
			freeimage(ps.nsky);
		}

		/* Refine WCS for stack ?? */
		smooth = omega_smooth_image(ps.combined[0], 1);

		if(cpl_detector_interpolate_rejected(smooth) != CPL_ERROR_NONE)
			cpl_msg_debug(cpl_func,"Cannot clean rejected pixels. %s",
					cpl_error_get_message());

		xlist = cpl_propertylist_load(cpl_frame_get_filename(firstframe),j);

		/* Get filter information */
		plist_dummy = cpl_propertylist_load_regexp(
				cpl_frame_get_filename(firstframe),0,"ESO INS FILT",0);
		cpl_propertylist_copy_property_regexp(xlist,plist_dummy,"ESO INS FILT",0);
		freeplist(plist_dummy);




        /*Adjust CRPIX according to the CombMode used*/
	    par = cpl_parameterlist_find(pars, "omega.omega_science.CombMode");
	    sval = cpl_parameter_get_string(par);
	    if (!strcmp(sval, "union")){
		omega_correct_crpix_union(xlist, offsets);
	    }

		alist = cpl_propertylist_new();
		cpl_propertylist_append_string(alist, "EXTNAME",
				cpl_propertylist_get_string(xlist, "EXTNAME"));
		cpl_propertylist_set_comment(alist,"EXTNAME", "Extension name");


		cpl_propertylist_save(xlist,smoothed,CPL_IO_CREATE);
		cpl_image_save(smooth,smoothed,BITPIX,xlist,CPL_IO_EXTEND);

		cpl_propertylist_save(xlist,smoothed_weight,CPL_IO_CREATE);
		cpl_image_save(ps.combined[1],smoothed_weight,BITPIX,xlist,CPL_IO_EXTEND);


		cpl_msg_info(cpl_func,"Refining WCS for stack image");
		ps.plist = omega_match_points(smoothed, smoothed_weight, ps.usnoa2, pars, j,
								&omega_sci_qc.nmatches, &omega_sci_qc.mratio);
		if(ps.plist == NULL){
			cpl_msg_warning(cpl_func,"Unable to refine WCS for stack image");
			ps.plist = cpl_propertylist_duplicate(xlist);
		}

		freeimage(smooth);
		freeplist(xlist);

		ps.qclist = cpl_propertylist_new();
		cpl_propertylist_update_int(ps.qclist, "ESO QC SCI MATCHES", omega_sci_qc.nmatches);
		cpl_propertylist_set_comment(ps.qclist, "ESO QC SCI MATCHES",
				"number matches between USNOA2 cat and stars");
		cpl_propertylist_update_double(ps.qclist, "ESO QC SCI RATIO MATCHES", omega_sci_qc.mratio);
		cpl_propertylist_set_comment(ps.qclist, "ESO QC SCI RATIO MATCHES",
				"ratio USNOA2 matches/detected stars");

		/* Save stack and confidence map */
		if(isfirst){
			ps.proname = cpl_sprintf("%s_%s.fits", INSTRUME,STACK_PROCATG);
			ps.confname = cpl_sprintf("%s_%s.fits", INSTRUME,CONF_PROCATG);
			product_frame = omega_product_frame(ps.proname, STACK_PROCATG, CPL_FRAME_TYPE_IMAGE);
			conf_frame = omega_product_frame(ps.confname,CONF_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

		cpl_msg_info(cpl_func,"Saving stack and confidence map");

		if(!cpl_propertylist_has(ps.plist,"EXTNAME")){
			cpl_propertylist_copy_property(ps.plist,alist,"EXTNAME");
		}
		if(omega_save_image(ps.combined[0],set,pars,ps.plist,ps.qclist,CPL_BPP_IEEE_FLOAT,
				           ps.proname,RECIPE,product_frame,NULL,isfirst) == -1){
			cpl_msg_error(cpl_func,"Cannot save product. %s", cpl_error_get_message());
			freebivector(offsets);
			freeframeset(proset);
			freeframeset(proset_mask);
			freeplist(alist);
			omega_science_tidy(0);
			return -1;
		}

		if(omega_save_image(ps.combined[1],set,pars,alist,NULL,CPL_BPP_16_SIGNED,
				           ps.confname,RECIPE,conf_frame,NULL,isfirst) == -1){
			cpl_msg_error(cpl_func,"Cannot save product. %s", cpl_error_get_message());
			freebivector(offsets);
			freeframeset(proset);
			freeframeset(proset_mask);
			freeplist(alist);
			omega_science_tidy(0);
			return -1;
		}

		freeplist(ps.qclist);
		freeplist(ps.plist);
		if(ps.combined!=NULL) {
		freeimage(ps.combined[0]);
		freeimage(ps.combined[1]);
		freespace(ps.combined);
		}

		/* Create catalogue from stack */
		ps.catalogue = omega_science_catalogue(ps.proname, ps.confname, zeropoint_final, pars);
		if(ps.catalogue == NULL){
			cpl_msg_error(cpl_func,"Unable to create catalogue of stack. %s", cpl_error_get_message());
			freeplist(alist);
			freebivector(offsets);
			freeframeset(proset);
			freeframeset(proset_mask);
            omega_science_tidy(0);
            return -1;
		}

		if(omega_basic_param(ps.catalogue, pixscalex, &Seeing, &ellipticity) == -1){
			cpl_msg_warning(cpl_func,"Cannot find stars with CLASS_STAR near 1 to calculate seeing");
	    }


		qclist = cpl_propertylist_new();
		cpl_propertylist_update_double(qclist, "ESO QC SCI SEEING", Seeing);
		cpl_propertylist_set_comment(qclist, "ESO QC SCI SEEING", "calculated seeing of science field");
		cpl_propertylist_update_double(qclist, "ESO QC SCI ELLIPTICITY", ellipticity);
		cpl_propertylist_set_comment(qclist, "ESO QC SCI ELLIPTICITY", "calculated ellipticity of science field");


		/* Save catalogue */
		if(isfirst){
			ps.catname = cpl_sprintf("%s_%s.fits", INSTRUME,CAT_PROCATG);
			catframe = omega_product_frame(ps.catname, CAT_PROCATG, CPL_FRAME_TYPE_TABLE);
		}
		cpl_msg_info(cpl_func,"Saving catalogue");
		if(omega_save_table(ps.catalogue,set,pars,alist,qclist,ps.catname,RECIPE,
				           catframe,NULL,isfirst) == -1){
			freebivector(offsets);
			freeframeset(proset);
			freeframeset(proset_mask);
			freeplist(alist);
			freeplist(qclist);
			omega_science_tidy(0);
			return -1;
		}

		freetable(ps.catalogue);
		freebivector(offsets);
		freeplist(alist);
		freeplist(qclist);
	}	/* next extension */

	cpl_msg_info(cpl_func,"Cleaning up");
	//freebivector(offsets);
	freeframeset(proset);
	freeframeset(proset_mask);
	freeplist(alist);
	freeplist(qclist);
	omega_science_tidy(0);

	return 0;

}



static int omega_science_retrieve_input_param(const cpl_parameterlist  *  parlist)
{

	const cpl_parameter *par = NULL;
    cpl_errorstate prestate = cpl_errorstate_get();

    /* Load common parameters */
	par = cpl_parameterlist_find_const(parlist, "omega.omega_science.ExtensionNumber") ;
	omega_science_config.extnum = cpl_parameter_get_int(par) ;
	par = cpl_parameterlist_find_const(parlist, "omega.omega_science.OverscanMethod") ;
	omega_science_config.oc = cpl_parameter_get_int(par) ;

	if(!cpl_errorstate_is_equal(prestate)){
		cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
		return -1;
	}

    return 0;
}

static int omega_science_load_calib(int ext)
{

	/* Load mandatory calibration frames */
	/* master bias */
	ps.mbias = cpl_image_load(cpl_frame_get_filename(ps.mbiasfr), CPL_TYPE_FLOAT, 0, ext);
	if(ps.mbias == NULL){
		cpl_msg_error(cpl_func,"Cannot load Master Bias. %s", cpl_error_get_message());
		return -1;
	}
	/* master flat */
	ps.mflat = cpl_image_load(cpl_frame_get_filename(ps.mflatfr),CPL_TYPE_FLOAT,0,ext);
	if (ps.mflat == NULL){
		cpl_msg_error(cpl_func,"Cannot load Master Flat. %s",cpl_error_get_message());
		return -1;
	}

	/* bad pixels map */
	ps.bpm = cpl_image_load(cpl_frame_get_filename(ps.bpmfr), CPL_TYPE_INT, 0, ext);
	if(ps.bpm == NULL){
		cpl_msg_error(cpl_func, "Cannot load BPM. %s", cpl_error_get_message());
		return -1;
	}

	/* Photometric table */
	ps.photom = cpl_table_load(cpl_frame_get_filename(ps.photomfr),ext,0);
	if(ps.photom == NULL){
		cpl_msg_warning(cpl_func,"Cannot load the zeropoint table. %s",cpl_error_get_message());
		cpl_msg_warning(cpl_func,"Using values defined in the recipe parameter");
	}

	/* Load optional calibration frames */
	/* night sky flat */
/*	if(ps.nskyfr != NULL)
		ps.nsky = cpl_image_load(cpl_frame_get_filename(ps.nskyfr), CPL_TYPE_FLOAT,0,ext);
*/
	/* illumination map */
	if(ps.illumfr != NULL)
		ps.illum = cpl_image_load(cpl_frame_get_filename(ps.illumfr),CPL_TYPE_FLOAT,0,ext);

	/* fringes map */
/*	if(ps.fringesfr != NULL)
		ps.fringes = cpl_image_load(cpl_frame_get_filename(ps.fringesfr), CPL_TYPE_FLOAT,0, ext);*/

	return 0;
}

static cpl_error_code omega_correct_crpix_union(cpl_propertylist *xlist,
		cpl_bivector *offsets){

	cpl_vector * offset_x=NULL;
	cpl_vector * offset_y=NULL;
	double       crpix1=0.;
	double       crpix2=0.;

	offset_x=cpl_bivector_get_x(offsets);
	offset_y=cpl_bivector_get_y(offsets);


    if (cpl_vector_get_max(offset_x)-cpl_vector_get_min(offset_x) > 3000 ||
    		cpl_vector_get_max(offset_y)-cpl_vector_get_min(offset_y) > 6000){
    	/* FIXME: The combination method <union> is overruled in omega_jitter_stack
    	 * and therefore the wcs adjustment is not needed - dirty solution! */
    	return cpl_error_get_code();
    }


	if(cpl_propertylist_has(xlist, "CRPIX1") &&
			cpl_propertylist_has(xlist, "CRPIX2")){
		crpix1=cpl_propertylist_get_double(xlist, "CRPIX1");
		crpix2=cpl_propertylist_get_double(xlist, "CRPIX2");

		cpl_propertylist_update_double(xlist,"CRPIX1",
				crpix1-cpl_vector_get_min(offset_x));

		cpl_propertylist_update_double(xlist,"CRPIX2",
				crpix2-cpl_vector_get_min(offset_y));

	}

	return cpl_error_get_code();


}

/* Initialize the pointers */
static void omega_science_init(void) {
	ps.bpmfr = NULL;
	ps.fringesfr = NULL;
	ps.usnoa2 = NULL;
	ps.nskyfr = NULL;
	ps.illumfr = NULL;
	ps.mbiasfr = NULL;
	ps.mflatfr = NULL;
	ps.photomfr = NULL;
	ps.ph = NULL;
	ps.eh = NULL;
	ps.scifits = NULL;
	ps.sfits = NULL;
	ps.combined = NULL;
	ps.catalogue = NULL;
	ps.scilist = NULL;
	ps.labels = NULL;
	ps.mbias = NULL;
	ps.mflat = NULL;
	ps.bpm = NULL;
	ps.photom = NULL;
	ps.proname = NULL;
	ps.confname = NULL;
	ps.catname = NULL;
	ps.nsky = NULL;
	ps.illum = NULL;
	ps.fringes = NULL;
	ps.qclist = NULL;
	ps.plist = NULL;
}

/*
 * Free any allocated memory
 */
static void omega_science_tidy(int level) {
	freeimage(ps.mbias);
	freeimage(ps.mflat);
	freeimage(ps.bpm);
	freeimage(ps.illum);
	freetable(ps.photom);
	freefits(ps.scifits);
	freeframeset(ps.scilist);
	if(level == 1)
		return;

	freeimage(ps.fringes);
	freeimage(ps.nsky);
	freefits(ps.sfits);
	if(ps.combined!=NULL) {
	freeimage(ps.combined[0]);
	freeimage(ps.combined[1]);
	freespace(ps.combined);
	}
	freetable(ps.catalogue);
	freeimage(ps.mflat);
	freespace(ps.labels);
	freespace(ps.proname);
	freespace(ps.confname);
	freespace(ps.catname);
	freeplist(ps.qclist);
	freeplist(ps.plist);
}

/**@}*/
